import os
import json
import logging
from contextlib import redirect_stdout
import hvac
import slixmpp

LOG = logging.getLogger(__name__)
LOG_FORMAT = "%(levelname)-8s %(message)s"
XMPP_TEST_FILE = "/tmp/xmpp-test.txt"
XMPP_ROOM_SUFFIX = "@conference.xmpp.tpage.io"
DEFAULT_XMPP_ROOM = "fixme"
DEFAULT_XMPP_MESSAGE = "test"
ROOM_LIST = ["security", "backups", "system", "devops", "climate", "pydl", "debug"]
MESSAGE_PREFIX = {
    "debug": "[DEBUG]: ",
    "info": "[INFO]: ",
    "warn": "[WARN]: ",
    "error": "[ERROR]: ",
    "critical": "[CRIT]: ",
    "sec": "[SEC]: ",
}

CONFIG_FILE = os.getenv("HHELPER_CONFIG", "/opt/hhelper.json")
VAULT_CACHE_FILE = "/tmp/vault_cache.json"

# Send a message (and optional file) via XMPP to a room
class XmppMessage(slixmpp.ClientXMPP):
    def __init__(self, room, filename, message):

        # Get and load configuration
        config = load_config()
        jid = config.get("XMPP_JID")
        password = config.get("XMPP_PASSWORD")
        self.sender = config.get("XMPP_SENDER")

        slixmpp.ClientXMPP.__init__(self, jid, password)

        self.room = room
        self.filename = filename
        self.message = message

        self.add_event_handler("session_start", self.start)
        self.add_event_handler("failed_auth", self.disconnect_on_login_fail)

    def start(self, event):
        self.send_presence()
        self.get_roster()

        self.plugin["xep_0045"].join_muc(self.room, self.sender, wait=True)

        self.send_message(mto=self.room, mbody=self.message, mtype="groupchat")

        if self.filename is not None:
            LOG.info("Uploading file: %s", self.filename)

            def timeout_callback(arg):
                raise TimeoutError("could not send message in time")

            url = self["xep_0363"].upload_file(
                self.filename, timeout=60, timeout_callback=timeout_callback
            )
            LOG.info("Uploaded file: %s", self.filename)

            LOG.info("Sending file to room: %s", self.room)
            msg = self.Message(sto=self.room, stype="groupchat")
            msg["body"] = url
            msg["oob"]["url"] = url
            msg.send()

        self.disconnect()

    def disconnect_on_login_fail(self, event):
        LOG.warning("Failed to login: %s", str(event))
        self.disconnect()


# Load config from file
def load_config():
    with open(CONFIG_FILE) as config_file:
        return json.load(config_file)


# Attempt to read the cache file and return it's contents
# If it doesn't exist, we create it and populate it with an empty JSON array
# Returns: config in JSON format, False on failure
def get_vault_cache():
    if os.path.isfile(VAULT_CACHE_FILE) and os.stat(VAULT_CACHE_FILE).st_size > 0:
        with open(VAULT_CACHE_FILE, "r") as cache_file:
            try:
                cache = json.load(cache_file)
                return cache
            except json.decoder.JSONDecodeError as error:
                LOG.error(str(error))
                return False
    else:
        # Init file with dummy JSON data so json.load() doesn't complain
        with open(VAULT_CACHE_FILE, "w") as new_cache_file:
            new_cache_file.write("{}")
            return "{}"


# Attempt to update the cache file with the given key/value
# Returns: True if we update the cache file, False on failure
def add_to_vault_cache(secret, value):
    with open(VAULT_CACHE_FILE, "r") as cache_file:
        try:
            cache = json.load(cache_file)
        except json.decoder.JSONDecodeError as error:
            LOG.error(str(error))
            return False

    cache.update({secret: value})

    with open(VAULT_CACHE_FILE, "w") as cache_file:
        try:
            json.dump(cache, cache_file)
        except json.decoder.JSONDecodeError as error:
            LOG.error(str(error))
            return False

    return True


# Checks the current token TTL and renews if necessary
# Returns: False if renewing the token fails, True otherwise (renewed/no action needed)
def check_and_renew_token(vault):

    # Check current token
    try:
        token_info = vault.lookup_token()
    except hvac.exceptions.Forbidden:
        message = "Forbidden (token probably expired)"
        LOG.error("%s", message)
        send_log_message("security", "sec", message)
        return False

    # Check expiration time for renewal (<6 hours)
    if (
        token_info.get("data").get("ttl")
        and int(token_info.get("data").get("ttl")) < 21600
    ):
        renewed_token = vault.renew_token()

        if not renewed_token:
            message = "Failed to renew token, TTL: " + str(
                token_info.get("data").get("ttl")
            )
            LOG.error("%s", message)
            send_log_message("security", "error", message)
            return False

    return True


# Test Vault server
# Test Criteria: initialized, unsealed, valid token, and successful list of secrets
# Returns: hvac.Client if all criteria passes, False if any don't
def vault_test(address, entity, token):

    # Init the client
    vault = hvac.Client(url=address, token=token)

    # If Vault is not initialized or sealed, it's not ready
    if not vault.sys.is_initialized() or vault.sys.is_sealed():
        message = "Vault is not initialized or is sealed"
        LOG.error("%s", message)
        send_log_message("system", "error", message)
        return False

    # Check/Renew token if necessary
    good_token = check_and_renew_token(vault)

    if not good_token:
        return False

    # Check if we are authenticated
    if not vault.is_authenticated():
        message = "Failed to authenticate to Vault"
        LOG.error("%s", message)
        send_log_message("system", "error", message)
        return False

    # Do a test read to confirm we're good
    if not vault.secrets.kv.list_secrets(mount_point=entity, path="/"):
        message = "Failed to list secrets at: " + entity
        LOG.error("%s", message)
        send_log_message("system", "error", message)
        return False

    # Testing is done if we get here, return the hvac.Client instance
    LOG.info("Connected to Vault")
    return vault


# Get an instance of XmppMessage
# Returns: a ready-to-connect instance of XmppMessage
def get_xmpp(room=DEFAULT_XMPP_ROOM, message=DEFAULT_XMPP_MESSAGE, file=None):

    xmpp = XmppMessage(room, file, message)
    xmpp.register_plugin("xep_0045")
    xmpp.register_plugin("xep_0030")
    xmpp.register_plugin("xep_0199")
    xmpp.register_plugin("xep_0066")
    xmpp.register_plugin("xep_0071")
    xmpp.register_plugin("xep_0128")
    xmpp.register_plugin("xep_0363")

    # Return XmppMessage instance
    return xmpp


# Get an instance of hvac.Client, this is a wrapper for load_config() + vault_test()
# Returns: authenticated hvac.Client instance, False on failure
def get_vault():

    # Get and load configuration
    config = load_config()
    vault_address = config.get("VAULT_ADDR")
    vault_entity = config.get("VAULT_ENTITY")
    vault_token = config.get("VAULT_TOKEN")

    # Test configuration
    vault = vault_test(vault_address, vault_entity, vault_token)

    if not vault:
        LOG.error("Vault is not ready")
        send_log_message("system", "error", "Vault is not ready")
        return False

    return vault


# Get a secret, with caching support
# Returns: secret's value, False on failure
def get_vault_secret(secret, cache=True):

    # Get vault
    vault = get_vault()

    # Check token regardless of using cache or not, to prevent it from expiring
    good_token = check_and_renew_token(vault)

    if not good_token:
        LOG.warning("Token is close to expiring or is expired!")

    if len(str(secret).split("/")) == 2:
        mount_point = str(secret).split("/")[0]
        path = str(secret).split("/")[1]
    else:
        LOG.error("Unrecognized secret format: %s", secret)
        return False

    # Check if cache file exists, and pull from there if possible
    if cache:
        cache_file = get_vault_cache()

        if path not in cache_file:
            return get_vault_secret(secret, cache=False)

        LOG.info("Using cached value for: %s", path)
        secret = cache_file.get(path)
    else:
        # Grab secret from vault
        vault = get_vault()

        if vault is False:
            return False

        # Read secret
        secret_lookup = vault.secrets.kv.v2.read_secret_version(
            mount_point=mount_point, path=path
        )
        secret_value = secret_lookup.get("data").get("data").get("secret")

        if not secret_value:
            LOG.error("Failed to get secret: %s", str(path))
            secret = False

        # Add to cache before returning it
        if add_to_vault_cache(path, secret_value):
            LOG.info("Added: %s to the cache file", str(path))
        else:
            LOG.warning("Failed to add: %s to the cache file", str(path))

        secret = secret_value

    return secret


# Send an XMPP message or file to a given room
def send_xmpp(room=DEFAULT_XMPP_ROOM, message=DEFAULT_XMPP_MESSAGE, file=None):
    xmpp = get_xmpp(str(room) + XMPP_ROOM_SUFFIX, message, file)

    if xmpp is False:
        LOG.error("Failed to get XMPP instance")
        return False

    xmpp.connect()
    xmpp.process(forever=False)
    return True


# Send a formatted log message via XMPP
def send_log_message(room, level, message):
    # Get correct MESSAGE PREFIX from dict
    prefix = MESSAGE_PREFIX.get(level, "FIXME: ")

    return send_xmpp(room=room, message=str(prefix + message))


# Print a formatted log message to a file
def log(message, level, log_file):
    # Get correct MESSAGE PREFIX from dict
    prefix = MESSAGE_PREFIX.get(level, "FIXME: ")

    with open(log_file, "a") as file:
        with redirect_stdout(file):
            print(str(prefix + message))


if __name__ == "__main__":
    # Do some basic sanity testing when the module is run standalone
    logging.basicConfig(level=logging.INFO, format=LOG_FORMAT)
    get_vault()

    with open(XMPP_TEST_FILE, "w") as test_file:
        test_file.write("test")
    send_xmpp(message="file", file=XMPP_TEST_FILE)
    os.remove(XMPP_TEST_FILE)
